import fs from "node:fs";
import { Environment, Transaction, Tree } from "@canvas-js/okra-node";
import { AbstractGossipLog } from "../AbstractGossipLog.js";
import { assert } from "../utils.js";
export class GossipLog extends AbstractGossipLog {
    static async open(init, path) {
        if (!fs.existsSync(path)) {
            fs.mkdirSync(path, { recursive: true });
        }
        const env = new Environment(path, { databases: 3 });
        const gossipLog = new GossipLog(env, init);
        await gossipLog.write(async (txn) => { });
        return gossipLog;
    }
    constructor(env, init) {
        super(init);
        this.env = env;
    }
    async close() {
        this.log("closing");
        await this.env.close();
    }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        const txn = new Transaction(this.env, { readOnly: true });
        try {
            const tree = new Tree(txn, { dbi: "messages" });
            for await (const node of tree.nodes(0, lowerBound ?? { key: null, inclusive: false }, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            txn.abort();
        }
    }
    async read(callback) {
        this.log("opening read-only transaction");
        return await this.env.read(async (txn) => {
            const headsDBI = txn.openDatabase("heads");
            const messagesDBI = txn.openDatabase("messages");
            const messages = new Tree(txn, { dbi: messagesDBI });
            return await callback({
                ancestors: this.indexAncestors ? GossipLog.getReadOnlyAPI(txn, txn.openDatabase("ancestors")) : undefined,
                messages,
                heads: {
                    get: (key) => txn.get(key, { dbi: headsDBI }),
                    entries: (lowerBound = null, upperBound = null, options = {}) => txn.entries(lowerBound, upperBound, { ...options, dbi: headsDBI }),
                },
            });
        });
    }
    async write(callback) {
        this.log("opening read-write transaction");
        return await this.env.write(async (txn) => {
            const messages = new Tree(txn, { dbi: txn.openDatabase("messages") });
            return await callback({
                messages,
                heads: GossipLog.getReadWriteAPI(txn, txn.openDatabase("heads")),
                ancestors: this.indexAncestors ? GossipLog.getReadWriteAPI(txn, txn.openDatabase("ancestors")) : undefined,
            });
        });
    }
}
GossipLog.getReadOnlyAPI = (txn, dbi) => ({
    get: (key) => txn.get(key, { dbi }),
    entries: (lowerBound = null, upperBound = null, options = {}) => txn.entries(lowerBound, upperBound, { ...options, dbi }),
});
GossipLog.getReadWriteAPI = (txn, dbi) => ({
    get: (key) => txn.get(key, { dbi }),
    set: (key, value) => txn.set(key, value, { dbi }),
    delete: (key) => txn.delete(key, { dbi }),
    entries: (lowerBound = null, upperBound = null, options = {}) => txn.entries(lowerBound, upperBound, { ...options, dbi }),
});
